/******************************************************************************
 *
 * descent.m
 *
 * date:   14/8/2001
 * author: Nils Bruin
 *
 * 2-descent and 2-isogeny descent on elliptic curves over general number fields
 *
 ******************************************************************************/

/******************************************************************************

Mathematical explanation

We consider an isogeny phi: E -> E'. We get the sequence

0 -> E[phi] -> E -> E' -> 0

By Galois-cohomology we get

0 -> E'(K)/phi(E(K)) -> H^1(K,E[phi]) -> H^1(K,E)

The image of E'(K)/phi(E(K)) inside H^1(K,E[phi]) are called the *trivial
homogeneous spaces* of E wrt. phi.

The Selmer group S^{phi}(E/K) consists of the homogeneous spaces that are
trivial everywhere locally, i.e. become trivial under the restriction maps
H^1(K,E[phi]) -> H^1(K_p,E[phi]), for all localisations K_p of K.

The size of S^{phi}(E(K)) gives an upper bound on the size of

E'(K)/phi(E(K))

and thus (by combining the same information for the dual isogeny) for the size
of E(K)/[degree(phi)]E(K) and thereby on the rank.

Currently, these functions are only implemented for (degree) 2-isogenies and the
multiplication-by-2-map.

Let E: y^2=x^3+a2*x^2+a4*x+a6=F(x) and let A= K[x]/F(x)

we have

H^1(K,E[2])=(A^*)/(A^*2)

For phi a 2-isogeny, we have

H^1(K,E[2])=(K^*)/(K^*2).

In each of these cases, elements of these quotient groups are represented by an
element from A or K resp.

This map is referred to as "Mu".

 ******************************************************************************/


declare attributes Map:Isogeny_Mu,locisgdata;

declare verbose EllSelmer,7;

/**
 ** IsogenyMu
 **
 ** gives the connecting homomorphism E'(K)->H^1(K,E[phi]). The map takes
 ** both points in E'(K) and x-coordinates (for which it isn't checked whether
 ** it is really an x-coordinate of a point). This is useful for local points.
 **
 ** It should be noted, however, that in the case where phi=[2], no provisions
 ** are taken to return a correct value for 2-torsion points that are not
 ** defined over K (how could it?)
 **
 ** This function has memory.
 **/

intrinsic IsogenyMu(phi::Map : Fields:={})->Map
  {returns the connecting homomorphism Codomain(phi)->H^1(K,Kernel(phi)).
  If phi is multiplication by 2, then the optional Fields parameter takes
  suggestions of fields that may be useful in the construction of the codomain.}
  if not(assigned phi`Isogeny_Mu) then
    E:=Codomain(phi);
    a:=aInvariants(E);
    R:=BaseRing(E);
    case Degree(phi):
    when 2:
      require a[1] eq 0 and a[3] eq 0:
        "Only implemented for curves of the form [0,a2,0,a4,a6]";
      linpol:=IsogenyMapPsi(DualIsogeny(phi));

      // This is the x-coordinate of the non-zero element of the kernel
      rt:=Roots(linpol)[1][1];
      
      //for this point, we have to compute mu(P) separately
      specval:=Evaluate(UnivariatePolynomial(E)/(Parent(linpol)![-rt,1]),rt);
      
      //Codomain(mu) can be represented by the base field for 2-isogeny
      A:=R;
      
      //the actual map:
      mumap:=function(P)
	if Type(P) eq PtEll then
          error if Parent(P) ne E, "Point must be in domain of isogeny";
          if P[3] eq 0 then
            return A!1;
          end if;
          P:=P[1]/P[3];
	end if;
	
	//the map is just(x,y):-> x-rt, except for the kernel,
	//where the specially computed value is returned
	if P eq rt then
	  return specval;
	else
	  return P-rt;
	end if;
      end function;
      
    when 4:
      require IsogenyIsMultiplicationByTwo(phi):"not implemented";
      require a[1] eq 0 and a[3] eq 0:
        "Only implemented for curves of the form [0,a2,0,a4,a6]";
      F:=UnivariatePolynomial(E);
      dF:=Derivative(F);
      
      //Construct the codomain and pass on "Fields" for construction.
      A:=quo<PolynomialRing(R)|F>;
      
      //iso is a map to the direct-sum-of-fields representation of A
      _,iso:=AbsoluteAlgebra(A:Fields:=Fields);
      
      //Pair the 2-torsion with their tentative images
      twotorx:=[<u[1],iso(u[1]-A.1)>:u in Roots(F)];

      //any non-multiplicative part needs to be replaced by an appropriate
      //value, which is the value of the derivate (which is non-zero due to
      //square-freeness. These values are mapped back into A for convenient
      //representation.
      twotorx:=[< u[1],<  (a eq 0) select 
                               Evaluate(dF,Parent(a)!u[1])
	                  else a: a in u[2]
	               >@@iso>: u in twotorx];
		       
      //After this prepwork, the actual def of the function is easy:
      mumap:=function(P)
        if Type(P) eq PtEll then
          error if Curve(P) ne E, "Point must be in domain of isogeny";
          if P[3] eq 0 then
            return A!1;
          end if;
          P:=P[1]/P[3];
	end if;
	
	//two-torsion get its precomputed value returned (note that twotorx
	//acts as an associative list)
	if exists(val){pair[2]:pair in twotorx|pair[1] eq P}then
	  return val;
	else
	
	  //and otherwise, just return x-theta (where F(theta)=0).
          return P-A.1;
	end if;
      end function;
    else
      require false:"not implemented yet";
    end case;
    
    //the map definition takes a cunning coproduct-construction to accept
    //both fully qualified points and just x-coordinates.
    phi`Isogeny_Mu:=map<cop<R,E>->A| u:->mumap(Retrieve(u))>;
  end if;
  return phi`Isogeny_Mu;
end intrinsic;

// The record format for data stored in Isogenies indexed by prime-ideals.
// Should anyone else want to use these routines, they're free to extend
// this record structure. However, to make sure there is no interference with
// other routines, the following calling sequence should be adhered to whenever
// someone wants to store information is, say field X.
//
// rc:=IsogenyLocalData(phi,p);
// rc`X:=<new value>;
// StoreIsogenyLocalData(phi,p,rc);
//
// Retrieving a value can just be done by
//
// Xval:=IsogenyLocalData(phi,p)`X

RFlocisgdata:=recformat<xlist,twotor,twotorimg>;

/**
 ** IsogenyLocalData
 **/

intrinsic IsogenyLocalData(phi::Map,p::RngOrdIdl)->Rec
  {lookup local data in an isogeny}

  if not(assigned(phi`locisgdata)) then
    phi`locisgdata:=[];
  end if;
  if exists(t){u[2]: u in phi`locisgdata | u[1] eq p} then
    return t;
  else
    return rec<RFlocisgdata|xlist:=[]>;
  end if;
end intrinsic;

/**
 ** StoreIsogenyLocalData
 **/
  
intrinsic StoreIsogenyLocalData(phi::Map,p::RngOrdIdl,data::Rec)
  {store local data in an isogeny}

  if not(assigned(phi`locisgdata)) then
    phi`locisgdata:=[];
  end if;
  if exists(j){i:i in [1..#phi`locisgdata] | phi`locisgdata[i][1] eq p} then
    phi`locisgdata[j]:=<p,data>;
  else
    Append(~phi`locisgdata,<p,data>);
  end if;
end intrinsic;

/**
 ** LocalSelmerImageOfTwoTorsion
 **
 ** This routine depends on LocalTwoSelmerMap having memory
 ** for LocalTwoSelmerMap(p) and LocalTwoSelmerMap(<RngUPolRes>,p)
 **
 ** It computes LocalTwoSelmerMap(Mu(E'[2](K)))
 ** (this is useful because this is not computable through Mu directly)
 **
 ** Since the computations are expensive, this routine has memory
 ** (making use of IsogenyLocalData`twotorimg)
 **/
 
intrinsic LocalSelmerImageOfTwoTorsion(phi::Map,p::RngOrdIdl)->GrpAb
  {determines the image of the twotorsion in the selmer group. Presently
  only implemented for 2-isogeny and multiplication by two and phi is assumed
  to be of that type.}

  if not(assigned IsogenyLocalData(phi,p)`twotorimg) then
    case Degree(phi):
    when 4:
      Kp,mp:=MyCompletion(p);
      mpi:=Inverse(mp);
      F:=UnivariatePolynomial(Codomain(phi));
      dF:=Derivative(F);
      //First determine 2-torsion over number field
      twotor:=[u[1]:u in Roots(F)];
      
      A:=Codomain(IsogenyMu(phi));
      Sum,iso:=AbsoluteAlgebra(A);
      msqmap,Comps:=LocalTwoSelmerMap(A,p);
      //for those points, the map IsogenyMu will just return the correct value
      //twotorimg is just a list of group elts for now
      twotorimg:=[msqmap(IsogenyMu(phi)(u)):u in twotor];
	
      //take out the roots we have already taken care of
      F:=Parent(F)!(F/&*[Parent(F)|Parent(F)![-l,1]:l in twotor]);
      
      //check if E'[2](Kp) is strictly bigger than E'[2](K)
      extras:=[mpi(u[1]):u in MyRoots(Polynomial(Kp,[mp(v):v in Eltseq(F)]))];

      //if it is the case, then do some cunning surgery to compute the image
      //of these points
      if #extras gt 0 then
        vprint EllSelmer:"There's extra 2-torsion at this prime.";
	
	//add the extra torsion
	twotor:=twotor cat extras;
	
	isotheta:=iso(A.1);
	
	//compute their images under IsogenyMu by hand
	for x in extras do

	  //this would be the result (note that none of them actually IS 0) 
	  av:=Sum!<x-thetai:thetai in isotheta>;
	  valvec:=[Valuation(av[tupple`i],tupple`p):tupple in Comps];
	  vprint EllSelmer:"For x=",x,"we find valuation vector=",valvec;

	  //find the prime at which this thing is close to zero
	  _,idx:=Maximum(valvec);

	  //this should be true, I guess ... However, root finding involves
	  //losing precision, so it doesn't work.
          //assert valvec[idx] ge Precision(Kp);
	  
	  //now replace the value at that prime with dF(x) (or rather
	  //isotheta, the thing that should be approximated)
	  //and while we're at it, map through the local 2-selmer map.
	  //concatenate and cast into the appropriate group.
	  imgav:=&cat[Eltseq(Comps[j]`map(
		                (j eq idx) select
	                          Evaluate(dF,isotheta[Comps[j]`i])
	                        else
	                          av[Comps[j]`i])):j in [1..#Comps]];
          Append(~twotorimg,Codomain(msqmap)!imgav);
	end for;				  
      end if;
      
      //take the group these points generate and we're done!
      twotorimg:=sub<Codomain(msqmap)|twotorimg>;
    when 2:
      Kp,mp:=MyCompletion(p);
      mpi:=Inverse(mp);
      F:=UnivariatePolynomial(Codomain(phi));
      
      //we prefer to represent two-torsion over the number-field by exact
      //x-coordinates
      twotor:=[u[1]:u in Roots(F)];

      //take out the part we've already taken care of
      F:=Parent(F)!(F/&*[Parent(F)![-l,1]:l in twotor]);
      
      //and throw in approximations of extra local 2-torsion
      vprint EllSelmer:"entering p-adic root finding (timing will display).";
      vtime EllSelmer: twotor:=twotor cat
         [mpi(u[1]):u in MyRoots(Polynomial(Kp,[mp(v):v in Eltseq(F)]))];
      msqmap:=LocalTwoSelmerMap(p);
      
      //IsogenyMu is just defined properly for all of these points
      //(because we've made sure that the kernel of phi' is represented
      //exactly)
      twotorimg:=sub<Codomain(msqmap)|[msqmap(mu(u)):u in twotor]>
        	where mu:=IsogenyMu(phi);
    else
      require false:"not implemented";
    end case;
 
    //the recommended storing procedure
    rc:=IsogenyLocalData(phi,p);
    rc`twotorimg:=twotorimg;
    rc`twotor:=twotor;
    StoreIsogenyLocalData(phi,p,rc);
  end if;
  return IsogenyLocalData(phi,p)`twotorimg;
end intrinsic;

/**
 ** TryLocalAddPoint
 **
 ** Takes an x, checks whether it is an x-coordinate of Codomain(phi) locally
 ** at prime p. If that is the case, it sees it it enlarges the image under
 ** IsogenyMu of the points already registered, together with the image of the
 ** two-torsion. If all of the criteria are met, then x is added to the "xlist"
 ** in the local info of phi at p and true is returned. Otherwise, the result is
 ** false.
 **/

function TryLocalAddPoint(phi,p,x)

  //retrieve the proper "mod squares map".
  //We must make sure we use a routine that has memory, because we
  //want to end up with the same map between calls.
  if Degree(phi) eq 2 then
    msqmap:=LocalTwoSelmerMap(p);
  else
    msqmap:=LocalTwoSelmerMap(Codomain(IsogenyMu(phi)),p);
  end if;

  //first check if "x" is x of 2-torsion, in which case the "point on curve"
  //test won't work. Those points are already taken care of anyway, so we
  //can safely turn down such a candidate anyway.
  //(by the time this routine is called, we can safely assume that
  //a1=a3=0)
  if x in IsogenyLocalData(phi,p)`twotor or
     msqmap(Evaluate(UnivariatePolynomial(Codomain(phi)),x)) ne
          Codomain(msqmap)!0 then
    return false;
  
  //Otherwise, test if this x-coordinate is already in the image of
  //the known points.
  elif msqmap(IsogenyMu(phi)(x)) in
       sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phi,p),
       [msqmap(IsogenyMu(phi)(u)): u in IsogenyLocalData(phi,p)`xlist]> then
    return false;
    
  //only if x gives new information, add it to xlist (otherwise, the list could
  //get VERY long)
  else
    vprint EllSelmer:"Added x-coordinate:",x;
    rc:=IsogenyLocalData(phi,p);
    Append(~rc`xlist,x);
    StoreIsogenyLocalData(phi,p,rc);
    return true;
  end if;
end function;

/**
 ** LocalSelmerGroup
 **/

intrinsic LocalSelmerGroup(phi::Map,p::RngOrdIdl:Trials:=100,Spread:=5)->GrpAb
  {Computes local Selmer group of an isogeny. The optional parameters specify
  how many random shots should be tried to get extra generations and the spread
  of the valuation of those points. Presently, there is no fall-back method for
  full 2-descent. In that case, random shooting is all that's tried.
  Note that, although computation might fail one time, any information found
  is stored in the isogeny. Simply calling the routine again might already
  lead to success.}
  
  //of course, we distinguish between the different types of isogenies again.
  case Degree(phi):
  when 2:
    
    // the total expected size is the size of the direct sum of the
    // local selmer group of phi and of its dual.
    TotalExpectedRank:=(2+InertiaDegree(p)*Valuation(Parent(p)!2,p));
    vprint EllSelmer:
      "Computing local 2-isg-Selmer-group at a prime above",Minimum(p),
      "[e,f]=",[RamificationDegree(p),InertiaDegree(p)];
    vprint EllSelmer:"Rank(S^{phi})+Rank(S^{phi'})=",TotalExpectedRank;

    phid:=DualIsogeny(phi);
    msqmap:=LocalTwoSelmerMap(p);
    msqmapi:=Inverse(msqmap);
    A:=Codomain(msqmap);
    
    //first, compute the images of the already known points

    selm:=sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phi,p),
      [msqmap(IsogenyMu(phi)(u)): u in IsogenyLocalData(phi,p)`xlist]>;
    selmd:=sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phid,p),
      [msqmap(IsogenyMu(phid)(u)):u in IsogenyLocalData(phid,p)`xlist]>;
      
    //if that already makes up for the expected size, we're done!
    //This happens if LocalSelmerGroup is called after success or if
    //the 2-torsion is already enough - often the case at odd primes.
    if Ngens(selm)+Ngens(selmd) ge TotalExpectedRank then
      vprint EllSelmer:"Succeeded using 2-torsion and stored points";
      return selm;
    end if;
    
    //generate random elements
    vprint EllSelmer:"Shooting for points [Trials,Spread]=",[Trials,Spread];
    //rather arbitrary bound, but it does reflect that at 2, you probably
    //need more digits.
    Bound:=Minimum(p)^(2*Valuation(Parent(p)!4,p)+2);
    for i in [1..Trials] do
    
      //not much justification can be given for this way of generating
      //candidates, except that it has a nice spread in its most significant
      //digit(s) and spreads well in its valuation. In practice, it seems
      //to perform well for most curves.
      x:=Random(BaseRing(Codomain(phi)):Bound:=Bound)*
           SafeUniformiser(p)^Random([-Spread..Spread]);
	   
      //if adding the point succeeds, then check if we're finished
      if TryLocalAddPoint(phi,p,x) then
        selm:=sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phi,p),
        [msqmap(IsogenyMu(phi)(u)): u in IsogenyLocalData(phi,p)`xlist]>;
        if Ngens(selm)+Ngens(selmd) ge TotalExpectedRank then
	  vprint EllSelmer:"Succeeded using random points";
          return selm;
        end if;
      end if;
      
      //it's just a random point. Why not try the same for the dual isogeny?
      if TryLocalAddPoint(phid,p,x) then
        selmd:=sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phid,p),
        [msqmap(IsogenyMu(phid)(u)):u in IsogenyLocalData(phid,p)`xlist]>;
        if Ngens(selm)+Ngens(selmd) ge TotalExpectedRank then
	  vprint EllSelmer:"Succeeded using random points";
          return selm;
        end if;
      end if;
    end for;
    
    //By now, we've had enough and we mean serious business. We're just gonna
    //comb the homogeneous spaces.
    E:=Codomain(phi);
    Ed:=Domain(phi);
    kerpol:=IsogenyMapPsi(phid);
    P:=Parent(kerpol);
    t:=Roots(kerpol)[1][1];
    
    //The quadratic cofactor of the cubic defining the elliptic curve
    Q:=P!(UnivariatePolynomial(E)/P![-t,1]);
    
    kerpold:=IsogenyMapPsi(phi);
    Pd:=Parent(kerpold);
    td:=Roots(kerpold)[1][1];
    
    //and the same for the isogenous one
    Qd:=Pd!(UnivariatePolynomial(Ed)/Pd![-td,1]);
    
    //loop over all candidate images of the selmer group
    vprint EllSelmer:"Searching homogeneous spaces systematically.";
    vprint EllSelmer:"Ambient group is of size",#A;
    for a in A do
    
      //and be sure to NOT handle spaces we already know to be trivial
      if not(a in selm) then
      
        //take a representative 
        delta:=msqmapi(a);
	
	//and test the corresponding homogeneous space for local points.
        vprint EllSelmer:"trying homogeneous space for delta=",delta;
        //since we've already taken into account the 2-torsion, we know
	//that the quartic won't have roots, so we switch off the terribly
	//slow root finding
        bl,x:=HasLocalSquareValue(delta*Evaluate (Q,delta*P.1^2+t),p:NoRoots);
	
	//if one's found, we're in business.
        if bl then
	  
	  //compute corresponding x on elliptic curve
	  x:=delta*x^2+t;
	  vprint EllSelmer:"found x=",x;
	  
	  //the preconditions demand that adding this point succeeds.
	  error if not(TryLocalAddPoint(phi,p,x)),
	    "Local point found but cannot add it";
	    
	  //compute new, enlarged selmer group
          selm:=sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phi,p),
          [msqmap(IsogenyMu(phi)(u)): u in IsogenyLocalData(phi,p)`xlist]>;
	  
	  //and see if we're already done
          if Ngens(selm)+Ngens(selmd) ge TotalExpectedRank then
	    vprint EllSelmer:"Succeeded using systematic search";
            return selm;
          end if;
	else
          vprint EllSelmer:"no point found";
	end if;
      end if;
      
      //do the same thing for the dual. Note that there's a benefit to
      //doing them both in parallel (apart from that it saves writing a for
      //loop). It may be that selm has a large codimension and selmd has a
      //small one (or vice versa). If selm is in fact already known (twotor)
      //then this parallel strategy probably results in not enumerating all of
      //A, whereas if we were to do selm first, we would have to check an
      //enormous number of spaces, none of which yield anything.
      //(of course, this is just heuristics)
      if not(a in selmd) then
        delta:=msqmapi(a);
        vprint EllSelmer:"trying space of isogenous curve for delta=",delta;
        bl,x:=HasLocalSquareValue(delta*Evaluate(Qd,delta*Pd.1^2+td),p:NoRoots);
        if bl then
	  x:=delta*x^2+td;
	  vprint EllSelmer:"found x=",x;
	  error if not(TryLocalAddPoint(phid,p,x)),
 	    "Local point found but cannot add it";
          selmd:=sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phid,p),
          [msqmap(IsogenyMu(phid)(u)): u in IsogenyLocalData(phid,p)`xlist]>;
          if Ngens(selm)+Ngens(selmd) ge TotalExpectedRank then
	    vprint EllSelmer:"Succeeded using systematic search";
            return selm;
          end if;
        else
          vprint EllSelmer:"no point found";
	end if;
      end if;      
    end for;
    //since the systematic search procedure should be fail-safe, it is an error
    //to arrive here.
    error "Failed to determine the local Selmer group";
  when 4:
    vprint EllSelmer:
      "Computing local Mul-by-2-Selmer-group at a prime above",Minimum(p),
      "[e,f]=",[RamificationDegree(p),InertiaDegree(p)];

    //Make sure it's a mult-by-two map, not a 4-isogeny.
    require IsogenyIsMultiplicationByTwo(phi): "not implemented";

    F:=UnivariatePolynomial(Codomain(phi));

    //trigger computation of twotor and twotorimg (we need the size of twotor
    //to compute TotalExpectedSize)
    _:=LocalSelmerImageOfTwoTorsion(phi,p);
    TotalExpectedRank:=
      case<#IsogenyLocalData(phi,p)`twotor|
        0:0,1:1,3:2,default:false>+(InertiaDegree(p)*Valuation(Parent(p)!2,p));
    vprint EllSelmer:"Rank of local selmer group is",TotalExpectedRank;
    
    msqmap:=LocalTwoSelmerMap(Codomain(IsogenyMu(phi)),p);
    msqmapi:=Inverse(msqmap);

    //compute image of known points
    selm:=sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phi,p),
      [msqmap(IsogenyMu(phi)(u)):u in IsogenyLocalData(phi,p)`xlist]>;

    //this might already be enough
    if Ngens(selm) ge TotalExpectedRank then
      vprint EllSelmer:"Succeeded using 2-torsion and stored points";
      return selm;
    end if;
    
    vprint EllSelmer:"Trying 4-torsion ...";
    _,Div4Pol:=DivisionPolynomial(Codomain(phi),4);
    Kp,mp:=MyCompletion(p);
    extras:=[u[1]@@mp:u in
       MyRoots(Polynomial(Kp,[mp(v):v in Eltseq(Div4Pol)]))];
    if &or[TryLocalAddPoint(phi,p,u):u in extras] then
      if Ngens(selm) ge TotalExpectedRank then
	vprint EllSelmer:"Succeeded using 4-torsion";
	return selm;
      end if;
    end if;
    vprint EllSelmer:"Shooting for points [Trials,Spread]=",[Trials,Spread];

    //try random trials
    Bound:=Minimum(p)^(2*Valuation(Parent(p)!4,p)+2);
    for i in [1..Trials] do

      //we won't really hit all values with this, but it appears to be
      //sufficient
      x:=Random(BaseRing(Codomain(phi)):Bound:=Bound)*
	   SafeUniformiser(p)^Random([-Spread..Spread]);

      //if adding the point succeeds, test if we're done
      if TryLocalAddPoint(phi,p,x) then
        selm:=sub<Codomain(msqmap)|LocalSelmerImageOfTwoTorsion(phi,p),
          [msqmap(IsogenyMu(phi)(u)):u in IsogenyLocalData(phi,p)`xlist]>;
        if Ngens(selm) ge TotalExpectedRank then
	  vprint EllSelmer:"Succeeded using random points";
          return selm;
	end if;
      end if;
    end for;
    
    //had enough. Systematic search to be implemented here. First get good
    //rep of homogeneous spaces.
    vprint EllSelmer:"Giving up on random trials.";
    error "giving up. Run again.";
  else
    require false:"not implemented";    
  end case;
end intrinsic;

/**
 ** SelmerGroup
 **/
 
intrinsic SelmerGroup(phi::Map)->GrpAb,Map
  {Computes the selmer group of an isogeny. The second returned value is a map
  from the codomain of IsogenyMu(phi) to the ambient group of the selmer group}
  E:=Codomain(phi);
  require EllIsIntegral(E):
    "elliptic curve should have integral coefficients";
  case Degree(phi):
  when 2:
    E:=Codomain(phi);
    rt:=Roots(IsogenyMapPsi(DualIsogeny(phi)))[1][1];
    vprint EllSelmer:"Computing 2-isogeny-Selmer group for",E,
      "with x(kernel)=",rt;   
  
    OK:=IntegerRing(BaseRing(E));
    
    //determine primes of bad reduction. We only need to consider primes
    //that multiply divide the discriminant (because otherwise, the Neron model
    //has one component anyway), and the primes above 2 of course
    S:={ LookupPrime(p[1]):p in Factorisation(2*Disc*OK)|
         (Minimum(p[1]) eq 2) or Valuation(Disc,p[1]) gt 1}
	 where Disc:=Discriminant(E);
    vprint EllSelmer:"Really bad primes are",S;
    vprint EllSelmer:"Computing 2-selmer group of",OK:Minimal;
    slmAmb,mp:=pSelmerGroup(OK,2,S);
    imp:=Inverse(mp);
    slm:=slmAmb;
    F:=UnivariatePolynomial(E);
    F:=Parent(F)!(F/Polynomial(Parent(rt),[-rt,1]));
    B,A:=Explode(Eltseq(F));
    vprint EllSelmer:"A priori bound on 2-rank of Selmer group:",Ngens(slm);
    
    //first handle the infinite primes
    for i in [1..Signature(OK)] do

      //only under these condition, we'll be getting info from this prime
      if Sign(Conjugate(A^2-4*B,i)) lt 0 or
        ( Sign(Conjugate(2*rt+A,i)) lt 0 and
	  Sign(Conjugate((2*rt+A)^2-(A^2-4*B),i)) gt 0) then
	  
	//define image group (will be the "sign")
	gp:=AbelianGroup([2]);
	
	//compute "sign" map on ambient selmer group
	lsm:= hom< slmAmb->gp|
	  [((Conjugate(imp(a),i) lt 0) select 1 else 0)*gp.1 :
	     a in OrderedGenerators(slmAmb)] >;
	//by conditions above, we need to be in the kernel of this map.
        slm:= slm meet Kernel(lsm);
        vprint EllSelmer:"Bound after real embedding nr.",i,"is",Ngens(slm);
      else
        vprint EllSelmer:"Real embedding nr.",i,"does not give information.";
      end if;
    end for;
    //now loop through bad primes
    for p in S do
      msqm:=LocalTwoSelmerMap(p);
      lslm:=LocalSelmerGroup(phi,p);
      
      //quotient out by local selmer group
      quogrp,quomap:=quo<Codomain(msqm)|lslm>;
      
      //compose all maps into a homomorphism from ambient group to quotient
      lsm:=hom<slmAmb->quogrp |
         [quomap(msqm(imp(a))):a in OrderedGenerators(slmAmb)]>;
      quogrp,quomap:=quo<Codomain(msqm)|lslm>;
      
      //and, of course, we have to be in the kernel of it.
      slm:=slm meet Kernel(lsm);
      vprint EllSelmer:"Bound on global selmer group is now",Ngens(slm);
      //in fact, we could break here if #slm meets a lower bound (1?)
    end for;
    vprint EllSelmer:"Finished with global selmer group.";
    return slm,mp;
  when 4:
    require IsogenyIsMultiplicationByTwo(phi): "not implemented";
    E:=Codomain(phi);
    vprint EllSelmer:"Computing Mul-by-2-Selmer group for",E:Minimal;
   
    require EllIsIntegral(E):
      "elliptic curve should have integral coefficients";
    OK:=IntegerRing(BaseRing(E));
    
    //bad primes. exclude odd ones where disc has valuation 1
    S:={ LookupPrime(p[1]):p in Factorisation(2*Discriminant(E)*OK) |
         (Minimum(p[1]) eq 2) or (Valuation(Disc,p[1]) gt 1) }
       where Disc:=Discriminant(E);
    vprint EllSelmer:"Really bad primes are",S;
    
    //the "ambient" group for the selmer group
    vprint EllSelmer:"Computing field decomposition of Algebra A=K[x]/F(x).";
    A:=Codomain(IsogenyMu(phi));
    vprint EllSelmer:"Computing 2-Selmer group of",A;
    slmAmb,mp:=pSelmerGroup(A,2,S);
    vprint EllSelmer:"A priori bound on 2-rank of Selmer group:",Ngens(slmAmb);
    
    //we need to take the "norm=square" part of that. We assume that
    //the norm of the "S-units" in A is an "S-unit" in OK.
    //this might be incorrect, because "S-unit" here may be enlarged to be able
    //to represent a part of the class group. So far, no errors occurred.
    //a possible mend would be: take representatives of slmAmb, take their
    //norms and take the set of primes occurring in those.
    //(so we initialise the slm as the kernel of the norm map)
    
    vprint EllSelmer:"Computing 2-selmer group of base field";
    slmK,mpK:=pSelmerGroup(OK,2,S);
    slm:=Kernel( hom<slmAmb -> slmK |
        [mpK(Norm(g@@mp)):g in OrderedGenerators(slmAmb)]>);
    vprint EllSelmer:"Bound after Norm criterion:",Ngens(slm);

    //the infinite primes!
    F:=UnivariatePolynomial(E);
    for i in [1..Signature(OK)] do
      Fr:=[Conjugate(f,i):f in Eltseq(F)];
      Rx:=PolynomialRing(Universe(Fr));
      Fr:=Rx!Fr;
      
      //Little problem in Magma: Even though Fr is a *REAL* polynomial,
      //roots still returns complex as well as imaginary roots.
      //the following should extract the real ones. Testing the
      //imaginary part for "eq 0" should work fine, but just to be sure, we've
      //included a few assertions as well.
      thetar:=[u[1]:u in Roots(Fr)];
      thetai:=[u:u in thetar| Im(u) ne 0];
      thetar:=[u:u in thetar| Im(u) eq 0];
      assert #thetai+#thetar eq 3;
      print thetai,thetar;
      assert #thetai eq 0 or (
        (#thetai eq 2) and 
        Abs((Im(thetai[1])+Im(thetai[2]))/Im(thetai[1])) lt 1e-20 and
	Abs((Re(thetai[1])-Re(thetai[2]))/Re(thetai[1])) lt 1e-20);
      //we should pick the left-most real zero. (everything should be on the
      //right of that, or: x-thetar >= 0.
      thetar:=Minimum(thetar);
      gp:=AbelianGroup([2]);
      
      //we have to be able to see the sign of the S-units. We take a naive
      //approach to computing it: We take real approximations of the S-units
      //wrt. to the power basis and we take the appropriate image of thetar.
      //Since S-units can be horrible wrt the power basis, we may be wasting
      //a lot of precision here. But precision is cheap :-)
      
      //However, MAGMA's precision tracking cannot be trusted. We
      //approach the problem naively: we trust that Claus did a good job on
      //Conjugate, so we assume the precision returned there is indeed
      //correct. Multiplying doesn't really hurt precision, so for elements
      //a0+theta*a1+theta^2*a2, we first compute
      //[conj(a0,i),thetar*conj(a1,i),thetar^2*conj(a2,i)]
      //higher entries may be missing, according to #Eltseq
      RealImages:=
        [[Conjugate(b[j],i)*thetar^(j-1):j in [1..#b]]
	    where b:=Eltseq(a@@mp): a in OrderedGenerators(slmAmb)];
      //Precs is (for each element) the exponent of the least significant
      //known digit of the RealImages. If a number falls below that, then we
      //know we have lost all precision in this number.
      Precs:=[Maximum([exp-Precision(c) where _,exp:=MantissaExponent(c):
        c in b]):b in RealImages];
      //We just compute the RealImages now, which means summing the sublists
      RealImages:=[&+b:b in RealImages];	
	
      //test if we can "reliably" tell the sign. We still want to have 3 digits
      //left, just to be safe.
      error if exists{i:i in [1..#Precs] |
        exp lt Precs[i]+3 where _,exp:=MantissaExponent(RealImages[i])},
	"Loss of precision. Cannot determine sign of S-units";

      //construct the appropriate sign map
      lsm:=hom<slmAmb->gp|[(b lt 0 select 1 else 0)*gp.1:b in RealImages]>;
      
      //and, of course, we want to be in the kernel of it. 
      slm:=slm meet Kernel(lsm);
      vprint EllSelmer:"Bound after real embedding nr.",i,"is",Ngens(slm);
    end for;
    
    //intersection at finite primes is completely straightforward
    //(and precise!)  
    for p in S do
      msqm:=LocalTwoSelmerMap(A,p);
      lslm:=LocalSelmerGroup(phi,p);
      quogrp,quomap:=quo<Codomain(msqm)|lslm>;
      lsm:=hom<slmAmb->quogrp |
         [quomap(msqm(a@@mp)):a in OrderedGenerators(slmAmb)]>;
      slm:=slm meet Kernel(lsm);
      vprint EllSelmer:"Bound on global selmer selmer group is now",Ngens(slm);
    end for;
    vprint EllSelmer:"Finished with global selmer group.";
    return slm,mp;      
  else
    error "not implemented";
  end case;
end intrinsic;

/**
 ** Rankbound(isogeny)
 **
 ** Just a wrapper around SelmerGroup. It interprets the info for you. 
 **/

intrinsic RankBound(phi::Map) -> RngIntElt
  {Returns a rank bound on the domain and codomain of phi, computed using a
  phi-isogeny descent}
  case Degree(phi):
  when 2:
    return Ngens(SelmerGroup(phi))+Ngens(SelmerGroup(DualIsogeny(phi)))-2;
  when 4:
    return Ngens(SelmerGroup(phi))-
      #Factorisation(UnivariatePolynomial(Codomain(phi)))+1;
  else
    require false:"Only 2-isogeny and 2-descent implemented";
  end case;
end intrinsic;

/**
 ** RankBound(CrvEll)
 **
 ** Figures out for itself what kind of descents apply to the curve and
 ** does those and takes the minimum of the rank bounds thus obtained.
 ** By default, it does not do a full 2-descent if 2-isogenies are available.
 ** If the flag TwoDescent is set, it does. (a 2-descent can be considerably
 ** more expensive)
 **
 ** A big drawback of this function is that information is not cached, as
 ** the isogenies are created dynamically.
 **
 ** A further feature is, that it base changes an elliptic curve over
 ** Rationals() to a linear extension of Rationals(),
 ** so this function does apply to rational elliptic curves too.
 **/

intrinsic RankBound(E::CrvEll:TwoDescent:=false) -> RngIntElt
  {Compute a rank bound on the curve E using all available 2-isogenies}
  if Type(BaseRing(E)) eq FldRat then
    E:=BaseChange(E,NumberField(PolynomialRing(Rationals()).1-1:
                             DoLinearExtension:=true));
  end if;
  TwoTorsion:=[E![x[1],0,1]:x in Roots(UnivariatePolynomial(E))];
  if #TwoTorsion gt 0 then
    print "computing 2-isogeny rank bounds for points ",TwoTorsion;
    Bounds:=[RankBound(TwoIsogeny(P)):P in TwoTorsion];
    if TwoDescent then
      print "also computing 2-descent bound";
      Append(~Bounds,RankBound(MultiplicationByMMap(E,2)));
    end if;
    print "Bounds computed:", Bounds;
    return m where m:=Minimum(Bounds);
  else
    print "computing 2-descent bound";
    Bound:=RankBound(MultiplicationByMMap(E,2));
    return Bound;
  end if;
end intrinsic;
