/******************************************************************************
 *
 * etalealg.m
 *
 * date:   26/6/2001
 * author: Nils Bruin
 *
 * Some routines to work with algebras of the form A=K[x]/F(x), where K is a
 * number field and F is a square free polynomial.
 *
 ******************************************************************************/

declare attributes RngUPolRes: AbsoluteMap,LocalData;

declare verbose EtaleAlg, 1;

RFLocalData:=recformat<TwoSelmerMap,TwoSelmerMapComponents>;

/**
 ** AbsAlgLocalData
 **
 ** retrieve local data from an algebra
 **/

intrinsic AbsAlgLocalData(A::RngUPolRes,p::RngOrdIdl)->Rec
  {Retrieves local data from an algebra}

  if not(assigned A`LocalData) then
    A`LocalData:=[];
  end if;
  if exists(R){r[2]:r in A`LocalData| r[1] eq p} then
    return R;
  else
    R:=rec<RFLocalData|>;
    Append(~A`LocalData,<p,R>);
    return R;
  end if;
end intrinsic;

/**
 ** AbsAlgStoreLocalData
 **
 ** store local data in an algebra
 **/
 
intrinsic AbsAlgStoreLocalData(A::RngUPolRes,p::RngOrdIdl,R::Rec)
  {stores local data in an algebra}
  if not(assigned A`LocalData) then
    A`LocalData:=[];
  end if;
  if exists(i){i: i in [1..#A`LocalData]| A`LocalData[i][1] eq p} then
    A`LocalData[i][2]:=R;
  else
    Append(~A`LocalData,<p,R>);
  end if;
end intrinsic;    

/**t
 ** AbsoluteAlgebra
 **
 ** The biggie! This routine decomposes an algebra into field extensions of its
 ** base field. It represents those fields as absolute extensions.
 ** It takes an optional list of suggested fields. If it finds it needs a field
 ** isomorphic to one of the suggested fields, it will use that one.
 **/
/*************************
function MyIsIsomorphic(N,M)
  if AbsoluteDegree(N) eq 1 and AbsoluteDegree(M) eq 1 then
    return true,map<N->M|a:->M!Norm(a),b:->N!Norm(b)>;
  else
    return a,b where a,b:=IsIsomorphic(N,M);
  end if;
end function;
*************************/
intrinsic AbsoluteAlgebra(A::RngUPolRes:Fields:={})->SetCart,Map
  {Given a semi-simple algebra over a number field, returns the isomorphic
  direct sum of absolute fields as a cartesian product and the isomorphisms
  to and from this product. The optional parameter Fields enables the user to
  suggest representations of the absolute fields. Af this function finds it
  needs a field isomorphic to one occurring in the supplied list, then it will
  use the given field. Otherwise it will construct a field itself.
  The isomorphism is returned as a second argument. If called twice on the
  same algebra, it will recompute if the Fields argument is given. Otherwise
  it will return the result computed the first time.}
  
  // Computations here are expensive and what we compute is canonical,
  // so we store the info in an attribute. Note that when this routine is
  // called with "Fields" set to some non-empty set, then the user probably has
  // a good reason for wanting to use those fields, so then we recompute too
  // (and store the result for future reference)

  if Fields ne {} or not(assigned(A`AbsoluteMap)) then
    F:=Modulus(A);
    vprint EtaleAlg: "Computing field decomposition of",A;
    require Discriminant(F) ne 0:
      "Algebra should be semisimple";
      
    //determine the irreducible components of A (there are "r" of them)
    fct:=[f[1]: f in Factorisation(F)];
    vprint EtaleAlg: "Modulus factors as",fct;
    r:=#fct;
    K:=BaseRing(A);
    
    //due to current restrictions, the ground field of A should be an absolute
    //number field (yes *number field*. Rationals won't do)
    require (Type(K) eq FldNum) and IsAbsoluteField(K):
      "Should be an algebra over an absolute number field";
      
    //irreducible components represented as a relative extension of K
    RelFld:=[NumberField(f:DoLinearExtension:=true): f in fct];
    
    //for absolute representations, try to use presupplied fields. This
    //allows precomputed classgroup information to be used in future
    //computations. Otherwise, for degree 1 extensions just take the ground
    //field. In the rest of the situations, take an OptimisedRepresentation
    //of an absolute field.
    AbsFld:=[ exists(Labs){L:L in Fields|IsIsomorphic(L,Lrel)} select
      Labs else 
	(Degree(Lrel) eq 1) select
           K else OptimisedRepresentation(AbsoluteField(Lrel)) :
      Lrel in RelFld];
    
    //The algebra is isomorphic to the cartesian product of these fields
    Sum:=CartesianProduct(AbsFld);
    vprint EtaleAlg:"field decomposition:",Sum;
    
    //Store the isomorphisms between the relative and absolute fields
    print RelFld,AbsFld;
    
    RelToAbsMap:=[phi where _,phi:=IsIsomorphic(RelFld[i],AbsFld[i]):
      i in [1..r]];
    //We compute the image of the power basis of A wrt the relative bases
    //of the irreducible components. Inverting this matrix gives the map
    //from the direct sum of the relative fields to A.
    M:=Matrix([ &cat[Eltseq(L.1^i):L in RelFld]:i in [0..Degree(F)-1]])^(-1);

    //Therefore, the isomorphism from A to Sum can be done by first
    //taking a polynomial representation of en element of A, evaluating that
    //in the image of the generator of A in each of the relative fields
    //(this is by construction the generator of the relative field)
    //and mapping through the isomorphism  to the absolute field
    //
    //The other way around is also simple: map from the absolute field
    //to the relative ones, take the coefficient sequences of these elements
    //and concatenate them. Multiply this vector with M and we have a coordinate
    //vector wrt the power basis of A. We simply coerce it into A.
    iso:=map< A->Sum |
       a:-> <RelToAbsMap[i](Evaluate(a,RelFld[i].1)): i in [1..r]>,
       tp:-> A!Eltseq(
	 Vector(&cat[Eltseq(tp[i]@@RelToAbsMap[i]): i in [1..r]])*
	 M)>;
    //store the map
    A`AbsoluteMap:=iso;
  end if;
  
  //perhaps this routine should JUST return a map? Returning the codomain
  //is a little superfluous.
  return Codomain(A`AbsoluteMap),A`AbsoluteMap;
end intrinsic;

intrinsic pSelmerGroupCheap(O::RngOrd,p::RngIntElt,S::SetEnum)->GrpAb,Map
  {Cheap p-selmer group. Does not compute inverse of map}
  require p eq 2:"Only 2-selmer groups implemented at the moment";
  require ClassNumber(O) mod 2 eq 1: "Only odd class number at the moment";
  sunits,sumap:=SUnitGroup(&*S);
  gn:=[sumap(u):u in OrderedGenerators(sunits)];
  slmgrp:=AbelianGroup([2: i in gn]);
  nofun:=function(a) error "not available in cheap version"; return 0; end
  function;
  slmmap:=hom<O->slmgrp| b:->nofun(b),a:->&*[gn[i]^Eltseq(a)[i]:i in [1..#gn]]>;
  return slmgrp,slmmap;
end intrinsic;

/**
 ** pSelmerGroup
 **
 ** Computes the p-SelmerGroup of A, i.e., the direct product of the
 ** pSelmerGroups of the direct summands of A. S should be a set of primes
 ** of the base field of A.
 **/

intrinsic pSelmerGroup(A::RngUPolRes,p::RngIntElt,S::{RngOrdIdl})->GrpAb,Map
  {returns the p-Selmer group of a semi-simple algebra. The list of
  ideals should be prime ideals of the underlying number field. The group
  returned is the direct sum of the selmer groups of the irreducible summands of
  A and the map is the obvious embedding in the multiplicative group of A.
  Note that, contrary to the version for fields, this routine takes an algebra
  over a field, not its ring of integers}
  
  vprint EtaleAlg:"Computing",p,"-Selmer group of",A,"with respect to",S;

  //retrieve direct product rep of the algebra
  Field,iso:=AbsoluteAlgebra(A);
  r:=NumberOfComponents(Field);

  //this is going to be a list of generators of the p-Selmer group
  generators:=[];
  
  //these will be the summand-wise maps of the p-Selmer groups
  mp:=[* *];
  
  //for all direct summands
  for i in [1..r] do
    vprint EtaleAlg:" Computing Selmer group for",Field[i];
    //let OL be the ring of integers of the i-th summand
    OL:=IntegerRing(Field[i]);
    
    //take the p-Selmer group wrt all primes above primes in S
    LS:=&join{Support(OL!!P) :P in S};
//    if #mp gt 0 then print Universe(mp); end if;
    grp,mp[i]:=pSelmerGroup(OL,p,LS);
    vprint EtaleAlg:" Computed a group with",Ngens(grp),"generators.";
    
    //for the generators of that p-Selmer group
    for g in OrderedGenerators(grp) do
      //build a tuple of all 1, except in the i-th entry,
      //where we put a representing element of the generator g.
      tp:=<>;
      for j in [1..r] do
	Append(~tp,Field[j]!((j eq i) select g@@mp[i] else 1));
      end for;
      //We take the representative in A of this tuple and put it in "generators"
      Append(~generators,tp@@iso);
    end for;
  end for;
  gens:=#generators;
  //build the appropriate abelian group
  grp:=AbelianGroup([p:i in [1..gens]]);
  //the p-Selmer map simply consists of projecting on each of the summands,
  //applying the selmer map for that summand and concatenating these results.
  //for the converse we can just multiply together the appropriate
  //representatives of the generators (which we computed above)
  grpmap:=map<A->grp|
    a:->grp!&cat[Eltseq(
        mp[i](Denominator(b)^2*b where b:=iso(a)[i])): i in [1..r]],
    g:->(&*[generators[i]^v[i]:i in [1..gens]] where v := Eltseq(g))>;
  vprint EtaleAlg:"Pieced together, this gives a group with",Ngens(grp),
     "generators";
  return grp,grpmap;
end intrinsic; 

/**
 ** Norm(RngUPolResElt)
 **
 ** You'd guess this routine would be present already ... It isn't!
 ** we just take the constant coefficient of the characteristic polynomial,
 ** with the appropriate sign.
 **/

intrinsic Norm(a::RngUPolResElt)->RngElt
  {Klugdy def for norm on polynomial residue rings}
  //The norm is just the constant coeff of the characteristic poly, with
  //a +- according to the parity of the degree. 
  return (-1)^(Degree(Modulus(Parent(a))))*
       Coefficient(CharacteristicPolynomial(a),0);
end intrinsic;

/**
 ** LocalTwoSelmerMap
 **
 ** The map itself is easy: simply concatenate the LocalTwoSelmerMaps of all
 ** the primes extending P of all the field-summands of A. The tricky part comes
 ** in the inverse, where some chinese remaindering must be done. Presently,
 ** applications of this routine (notably LocalSelmerGroup(<EllIsogeny>,p))
 ** DEPEND on the map being simply the concatenation of the maps occurring
 ** in the returned SeqEnum (in the order in which they occur).
 **
 ** this is partly due to the fact that DirectSum([<abelian groups>]) does not
 ** return inclusion & projection maps.
 **/

CRF:=recformat<i:RngIntElt,p:RngOrdIdl,map:Map>;

intrinsic LocalTwoSelmerMap(A::RngUPolRes,P::RngOrdIdl)->Map,SeqEnum
  {Direct product of the LocalTwoSelmerMaps of the direct summands of
  the semisimple algebra A. The second returned value is a sequence of
  tupples, consiting of <i,p,map>, where i is an index in the field
  decomposition of A, p is a prime extending P to the i-th field and map is
  LocalTwoSelmerMap(p). This information may be useful to twiddle around with
  yourself.}
  
  locdat:=AbsAlgLocalData(A,P);
  if not(assigned(locdat`TwoSelmerMap)) then

    Field,iso:=AbsoluteAlgebra(A);
    r:=NumberOfComponents(Field);

    //we store the map components. These correspond to the field decomposition
    //of A tensored with the completion of the base field at P.
    TwoSelmerMapComponents:=
       &cat [  [ rec<CRF|i:=j,p:=p,map:=LocalTwoSelmerMap(p)> : 
                         p in Support(IntegerRing(Field[j])!!P) ]:
              j in [1..r]];
    //The codomain is going to be the direct sum of the codomains of the
    //map components. However, MAGMA doesn't allow for automatic
    //construction of that. We do it by hand.
    
    grp:=AbelianGroup(&cat[Invariants(Codomain(tupple`map)):
       tupple in TwoSelmerMapComponents]);

    //One way, the map is easy. We take the representative in the field
    //decomposition, apply all map components (note that we have stored
    //the nr. of the field component of the ideal in tupple[1], so we know
    //what component to take of iso(a)) and concatenate the results
    //(and cast into the direct sum group)
    fn:=function(a)
	av:=iso(a);
	if exists{a:a in av|a eq 0} then
          error "ModSquaresMap only defined on multiplicative group";
	else
	  return grp!&cat[Eltseq(tupple`map(av[tupple`i])): 
	                        tupple in TwoSelmerMapComponents];
	end if;
      end function;
      
    //the inverse is quite a bit harder. We can take inverse images at
    //each of the primes, but we have to combine those by chinese remaindering
    //into a component. Finally we have to make sure that we compose the
    //element in A *multiplicatively*.
    
    fni:=function(b)
        b:=Eltseq(b);
	
	//we use that information of primes of the same field is adjacent
	//and that the information occurs in increasing order.
        refindex:=1;
	moduli:=[];
	elements:=[];
	av:=<>;
	n:=0;
        for tupple in TwoSelmerMapComponents do
	
	  //if the index of the field changes then we are finished
	  //with the present component. CRT and store it (and reset info) 
	  if tupple`i ne refindex then
	    assert tupple`i eq refindex+1;
	    Append(~av,CRT(elements,moduli));
	    refindex:=tupple`i;
	    moduli:=[];
	    elements:=[];
	  end if;
	  
	  //note that we allow ourselves some extra headroom to allow for
	  //the presence of a uniformiser. (val(4)+2 instead of val(4)+1).
	  Append(~moduli,tupple`p^(Valuation(Order(tupple`p)!4,tupple`p)+2));
	
	  GrpComponent:=Codomain(tupple`map);
	  
	  //advance indices to get the summand we should be dealing with
	  //(this has the effect of projecting on the direct summand of grp)
	  l:=n+Ngens(GrpComponent);
	  Append(~elements,Order(tupple`p)!
	     ((GrpComponent![b[k]:k in[n+1..l]])@@tupple`map));
	  n:=l;
        end for;
	
	//handle last component
	Append(~av,CRT(elements,moduli));
	
	//and compose element multiplicatively.
	return &*[<(i eq j) select
          av[i] else Field[j]!1 :j in [1..r]>@@iso: i in [1..r]];
      end function;
    locdat`TwoSelmerMap:=pmap<A->grp|a:->fn(a),b:->fni(b)>;
    locdat`TwoSelmerMapComponents:=TwoSelmerMapComponents;
    AbsAlgStoreLocalData(A,P,locdat);
  end if;
  return locdat`TwoSelmerMap,locdat`TwoSelmerMapComponents;
end intrinsic;
