/******************************************************************************
 *
 * locsol.m
 *
 * date:   26/6/2001
 * author: Nils Bruin
 *
 * Some routines to test equations for local solubility
 *
 ******************************************************************************/

/**
 ** HasSquareIntegralValues
 **/

intrinsic HasSquareIntegralValues(F::RngUPolElt :
   Neighbourhood:=false, NoRoots:=false)->BoolElt,RngLocElt
{Tests for integral solutions to y^2=F(x). The second return value
is an approximation of the x of such a point, if it exists. If the optional
parameter Neighbourhood is given, then only x that match up with Neighbourhood
(given the precision of Neighbourhood) are tried. If the flag NoRoots is
supplied, then F is not tested for having a p-adic root. Since p-adic
factorisation is slow, this saves some time, but may lead to very poor
performance and failure (loss of precision) if there is such a root}

  // first test if F has integral roots. Then 0^2=F(x) has a solution.
  // in fact, for this test, it's not even important that F is over a
  // local ring. (*SKIPPED IF NoRoots is true*)
  
  if not(NoRoots) then
    rts:=Roots(F);
    if #rts gt 0 then
      return true,rts[1][1];
    end if;
  end if;
    
  // now test the conditions 
  
  Op:=BaseRing(F);
  require Type(Op) eq RngLoc: "Polynomial needs to be over a local ring";
  if Neighbourhood cmpeq false then
    Neighbourhood:=elt<Op|0,0>;
  end if;
  require Parent(Neighbourhood) cmpeq Op:
    "Neighbourhood (if given) must be a member of the local ring";
  resfld:=ResidueClassField(Op);
  resfldreps:={@ elt<Op|u,1>: u in resfld @};
 
  if (Prime(Op) ne 2) then
    // this is the test-function for odd primes
    // this function takes an element (an origin) and
    // recursively looks in smaller neigbourhoods around it (while
    // perturbing it within its given precision)
    // to decide if F takes square values somewhere near origin.
    // if true is returned, then the second value gives a certificate
    // (i.e. an x for which F takes square values for all x+d within
    // the given precision of x)
    u:=elt<Op|UniformizingElement(Op),2>;
    test:=function(origin)
    
        // first, if we've really run out of precision, give an error
	
	if (Precision(Op) le Precision(origin)) then
          error "Insufficient precision for decision";
	end if;
	prc:=Precision(origin);

	//increase precision of origin in order to represent pertubations
	origin:=elt<Op|origin,prc+1>;
	//print "now working with origin=", origin;
	// loop through pertubations	
	for perturb in resfldreps do
          x:=origin+elt<Op|prc,perturb,prc+1>;
	  //print "trying x=",x;
	  y:=Evaluate(F,x);
	  vly:=Valuation(y);
	  if vly ge AbsolutePrecision(y) then
	  
	    // if y is indiscernable from 0 then recurse and look around
	    // current x.
	    pred,val:=$$(x);
	    if pred then
	      return pred,val;
	    end if;
	  elif ((vly mod 2) eq 0) and IsSquare(resfld!(y/u^vly)) then
	  
	    //otherwise, we can decide by looking at the parity of
	    //the valuation and the squareness of the first digit in the
	    //residue field.
	    return true,x;
	  end if;
	end for;

	//by now, we've tested everything and haven't found any square value
	//therefore, there are none.
	return false,_;
      end function;
  else
  
    // we construct the square units. That is, we take all units with precision
    // Valuation(2)+1 and square them. This gives us precision Valuation(4)+1
    // which is enough precision to tell square units from non-square ones.
    sqr:={
      elt<Op| &+[elt<Op|i-1,u[i],#u> : i in [1..#u]]^2,Valuation(Op!4)+1>  :
           u in Flat(car<{u : u in resfld | u ne 0},
                      CartesianPower(resfld,Valuation(Op!2))>)};
    v4p1:=Valuation(Op!4)+1;
    u:=elt<Op|UniformizingElement(Op),v4p1+1>;
    f:=Eltseq(F);
    test:=function(origin)
    
        print "origin=",origin;
        // first, if we've really run out of precision, give an error
	
	if (Precision(Op) lt AbsolutePrecision(origin)+v4p1) then
          error "Insufficient precision for decision";
	end if;
	prc:=AbsolutePrecision(origin);

	//increase precision of origin in order to represent pertubations
	origin:=elt<Op|origin,prc+1>;
	//print "now working with origin=", origin;
	// loop through pertubations	
	for perturb in resfldreps do
          x:=origin+elt<Op|prc,perturb,prc+1>;
	  //print "trying x=",x;
	  y:=&+[f[i]*(x^(i-1)):i in [1..#f]];
	  vly:=Valuation(y);
	  
	  // is there *ANY* information in y?
	  if RelativePrecision(y) gt 0 then
	  
	    // if y has odd valuation, there's no points around this x.
	    if (vly mod 2) eq 0 then
	    
	      // otherwise, multiply y into a unit (by a square!)
	      // and throw away any possibly excess precision.
	      yrd:=y/u^vly+elt<Op|v4p1,0,v4p1>;
	      
	      // now see if yrd is a square up to its precision
	      // (because we've thrown out excess precision, ALL digits
	      // of yrd should correspond to a listed square in sqr)
	      
	      if exists{sq: sq in sqr| RelativePrecision(yrd-sq) eq 0} then
	        // if yrd had full precision, then we're really done.
	        if AbsolutePrecision(yrd) ge v4p1 then
	          return true,x;
		
	        // Otherwise, look closer (i.e. recurse around this x)
	        else
	          pred,val:=$$(x);
	          if pred then
	            return pred,val;
	          end if;
		end if;
              end if;
	    end if;
	    
	  // if there was no info in y, look closer (recurse around this x).
	  else
	    pred,val:=$$(x);
	    if pred then
	      return pred,val;
	    end if;
          end if;
	end for;

 	//by now, we've tested everything and haven't found any square value
	//therefore, there are none.
	return false,_;
      end function;
   
  end if;
  a,b:=test(Neighbourhood);
  if a then
    return a,b;
  else
    return a,_;
  end if;
end intrinsic;

/**
 ** HasSquareValue
 **/
 
intrinsic HasSquareValue(F::RngUPolElt:NoRoots:=false)->BoolElt,RngLocElt
  {Tests whether y^2=F(x) has a solution in a local field. If so,
  the second value is an x of such a solution. Infinity() can also
  be returned. If NoRoots is supplied, then F is not tested for having
  roots. Since p-Adic factorisation is slow, this speed things up a bit, but
  if there is such a root, it might cause a loss of precision error.}
 
  Kp:=BaseRing(F); 
  require Type(BaseRing(F)) eq FldLoc: "Needs polynomial over local field";

  //If 0 is a root, things are nasty. Just return that root regardless
  //of NoRoots is set or not.
  f:=Coefficients(F);
  if f[1] eq 0 then
    return true,Kp!0;
  end if;
  
  //In the following, we use that Kp is a truly floating field, i.e., although
  //a priori the minimal polynomial of Kp may be only of finite precision,
  //precision is not truncated. That's very convenient here!
  
  //flatten out the newton polygon. (remember u for later use!);
  e:=Round((Valuation(f[#f])-Valuation(f[1]))/(#f-1));
  u:=UniformizingElement(Kp)^e;
  f:=[f[i]*u^(i-1):i in [1..#f]];

  //and shift the polygon above the axis, as close to the axis as possible.
  //(note that v will be even)
  v:=v-(v mod 2) where
         v:=Minimum([Valuation(fi):fi in f| fi ne 0]);
  //so we're multiplying f with a square here.	 
  f:=[fi/U: fi in f] where U:=UniformizingElement(Kp)^v;
  
  print [Valuation(fi):fi in f|fi ne 0];
  //reassemble F. If x is the original variable of F, then the F we have now is
  //the result of the substitution x=u*y (and multiplying F with a square).
  //we must remember to multiply any found value with u again to recover
  //the appropriate value for x again.
  F:=Parent(F)!f;

  // test for infinity and roots of F (if not inhibited)
  if not(NoRoots) then
    if (Degree(F) mod 2) eq 1 then
      return true,Infinity();
    end if;
    rt:=Roots(F);
    if #rt ne 0 then
      return true,rt[1][1]*u;
    end if;
  end if;
  
  F:=Polynomial(IntegerRing(BaseRing(F)),F);
  
  bl,x:=HasSquareIntegralValues(F:NoRoots);
  if bl then
    return bl,u*(Kp!x);
  end if;

  // next we reverse F, so that we can test values for which 1/x is
  // integral. If F is of odd degree, then shift the polynomial.
  // we're considering y^2=F(x) as a hyperelliptic curve, so this takes
  // care of the point "at infinity".
  F:=Polynomial(BaseRing(F),Reverse(Eltseq(F)))*Parent(F).1^(Degree(F) mod 2);
  print F, BaseRing(F);
  bl,x:=HasSquareIntegralValues(F :
         Neighbourhood:=O(UniformizingElement(BaseRing(F))),NoRoots);
  if bl then
    return bl,u/(Kp!x);
  end if;
  return false,_;
end intrinsic;

/**
 ** HasLocalSquareValue
 **/
intrinsic HasLocalSquareValue(F::RngUPolElt,p::RngOrdIdl:
     NoRoots:=false)->BoolElt,FldElt
  {Tests whether y^2=F(x) has a solution in the completion at p. The second
  value is an x of such a solution. Infinity() can also be returned}
  Kp,locmap:=MyCompletion(p);
  Fp:=Polynomial(Kp,[locmap(u):u in Eltseq(F)]);
  bool,x:=HasSquareValue(Fp:NoRoots:=NoRoots);
  if bool then
    print "Found: ",x;
    if x cmpeq Infinity() or Valuation(x) ge 0 then
      return bool,Inverse(locmap)(x);
    else
      return bool,1/(Inverse(locmap)(1/x));
    end if;
  else
    return false,_;
  end if;
end intrinsic;

/**
 ** IntegralRoots
 **/

intrinsic IntegralRoots(F::RngUPolElt) -> SeqEnum
  {Returns the integral roots of a square-free polynomial over a local field}
  R:=BaseRing(F);
  require Type(R) eq RngLoc: "Polynomial should be over a local ring";
  v:=Minimum([AbsolutePrecision(f):f in Eltseq(F)| f ne 0]);
  if v eq 0 then
    require false:"insufficient precision to separate roots";
  end if;
  Rbar,red:=ResidueClassField(R);
  Fbar:=Polynomial(Rbar,[red(f):f in Eltseq(F)]);
  dF:=Derivative(F);
  dFbar:=Derivative(Fbar);

  //note that any root over the integer ring should reduce to a root over the
  //finite field. We take representatives of these.
  RootsApprox:=[(u[1])@@red:u in Roots(Fbar)];
  Result:=[];
  for x0 in RootsApprox do
  
    //if the derivative in reduction is non-zero, then the derivative
    //at x0 is a unit. We can now use newton approximation (very fast)
    if Evaluate(dFbar,red(x0)) ne 0 then
      print "found separated root near",x0;
      x1:=x0;
      
      //repeat steps until value of polynomial at our root is
      //indistinguishable from 0. We can't do better than that.
      //also, this suggests that the precision in x1 is not exaggerated
      //provided that precision is kept track of properly.
      while RelativePrecision(Evaluate(F,x1)) gt 0 do
        x1:=x1-Evaluate(F,x1)/Evaluate(dF,x1);
	print "refining root to",x1;
      end while;
      
      //and store result.
      Append(~Result,<x1,1>);
    else
    
      //otherwise we have to look a bit closer. We take x0 as an approximation
      //and perturb with a polynomial variable. This is our new poly.
      G:=Evaluate(F,x0+UniformizingElement(R)*Parent(F).1);
      
      //of course, this polynomial has all of its coefficients 0 in reduction.
      //scalar multiplication does not change the roots, so just
      //scale down until things are visible in the finite field again.
      v:=Minimum([Valuation(g):g in Eltseq(G)]);
      print "looking around",x0,"zooming in with ",v;
      assert v gt 0;
      G:=Parent(G)![g/UniformizingElement(R)^v:g in Eltseq(G)];
      
      //we determine the integral roots of that G and we transform them
      //back into roots for F.
      Result cat:= [<x0+UniformizingElement(R)*z[1],1>:z in IntegralRoots(G)];
    end if;
  end for;
  return Result;
end intrinsic;

intrinsic MyRoots(F::RngUPolElt)->SeqEnum
  {Nils' own root finder - hopefully more efficient than the standard one}
  K:=BaseRing(F);
  require Type(K) eq FldLoc: "Base ring must be a local field";
  if Degree(F) eq 0 then
    require F ne 0: "roots not well-defined for a zero polynomial";
    return [];
  end if;
  O:=IntegerRing(K);
  f:=Eltseq(F);
  print(f);
  L:=[Valuation(fi):fi in f| fi cmpne Infinity()];
  n:=#L;
  //we transform x=u*z, so that F has only zeros for integral z.
  u:=UniformizingElement(K)^Minimum([Floor((L[i]-L[n])/(n-i)):i in [1..n-1]]);
  f:=[f[i]*u^(i-1):i in [1..n]];
  //furthermore, we rescale the poly to 0 content.
  c:=UniformizingElement(K)^Minimum([Valuation(fi):fi in f]);
  f:=Polynomial(O,[O!(fi/c):fi in f]);
  //transform back
  return [<x[1]*u,x[2]> : x in IntegralRoots(f)];
end intrinsic;
