/******************************************************************************
 *
 * mynumfld.m
 *
 * date:   26/6/2001
 * author: Nils Bruin
 *
 * Some features missing for number field and related objects
 *
 ******************************************************************************/

/* We'll be saving information in prime ideals later on. Therefore, it is good
to be able to get a "standard version" of a prime ideal.  With LookupPrime,
we store & retrieve standard versions of a prime ideal. Be sure to call

p:=LookupPrime(p);

whenever you want to use the functions here in order to get higher consistency
in returned values and a better performance */

declare attributes RngOrd: StandardPrimeIdeals;

intrinsic LookupPrime(P::RngOrdIdl) -> RngOrdIdl
  {Routine to get *IDENTICAL* prime ideals for a number field.
  Calling this routine several times with equal prime ideals of the same order
  is guaranteed to return identical prime ideals. This enables attributes to
  be stored in prime ideals and preserved along different instantiations
  of that prime ideal}
  O:=Order(P);
  if not(assigned O`StandardPrimeIdeals) then
    O`StandardPrimeIdeals:={};
  end if;
  if exists(Q){R:R in O`StandardPrimeIdeals| P eq R} then
    return Q;
  else
    Include(~O`StandardPrimeIdeals,P);
    return P;
  end if;
end intrinsic;

IdealRecordFormat:=recformat<LocalTwoSelmerMap,SafeUniformiser,
Completion,ComplMap>;

/**
 ** Completion with memory
 **/

intrinsic MyCompletion(p::RngOrdIdl:Precision:=false)->FldLoc,Map
  {Returns a completion with memory. If precision is given and is different from
  a stored completion, then stored field will be changed (recreated)}
  if not(assigned p`Record) then
    p`Record:=rec<IdealRecordFormat|>;
  end if;
  if not(assigned p`Record`Completion) or Precision cmpne false then
    if Precision cmpne false then
      Kp,mp:=Completion(NumberField(Order(p)),p:Precision:=Precision);
    else
      Kp,mp:=Completion(NumberField(Order(p)),p);
    end if;
    rc:=p`Record;
    rc`Completion:=Kp;
    rc`ComplMap:=map<Domain(mp)->Codomain(mp)|
          a:->mp(a*Denominator(a))/mp(Denominator(a)),
	  b:->Inverse(mp)(b)>;
    p`Record:=rc;
  end if;
  return p`Record`Completion,p`Record`ComplMap;
end intrinsic;

/**
 ** SafeUniformiser
 **
 ** retrieve or compute and store an element of valuation 1 at the given prime
 ** and valuation 0 at the other primes above the same rational prime.
 **/

intrinsic SafeUniformiser(p::RngOrdIdl)->RngOrdElt
  {returns an element of valuation 1 at the given prime and of
   valuation 0 at the other primes over Norm(p)}
   if not(assigned p`Record) then
     p`Record:=rec<IdealRecordFormat|>;
   end if;
   if not(assigned p`Record`SafeUniformiser) then
     I:=[q[1]: q in Factorisation(Norm(p)*Order(p))| q[1] ne p];
     u:=WeakApproximation([p]cat I,[1] cat [0: i in I]);
     rc:=p`Record;
     rc`SafeUniformiser:=u;
     p`Record:=rc;
   end if;
   return p`Record`SafeUniformiser;
end intrinsic;

/**
 ** LocalTwoSelmerMap
 **
 ** This routine returns a map from K^* -> Kp^* /(Kp^*2). If called twice with
 ** the same prime, returns the identical map. If called on equal but not
 ** identical primes, returns different maps that may not even agree
 ** coordinate-wise.
 **/
 
intrinsic LocalTwoSelmerMap(p::RngOrdIdl)->Map
  {returns a map from K^* to the group Kp^*/(Kp^*2)}
  if not(assigned p`Record) then
    p`Record:=rec<IdealRecordFormat|>;
  end if;
  if not(assigned p`Record`LocalTwoSelmerMap) then
    // We assume that grp1 is returned in Smith Normal Form!!!
    // therefore, by construction, we know that grp1 has all its
    // factors of even order.
    grp1,mp1i:=RayResidueRing(p^(Valuation(Order(p)!4,p)+1));
    grp2,mp2:=quo<grp1|[2*grp1.i: i in [1..Ngens(grp1)]]>;
    mp1:=Inverse(mp1i);
    mp2i:=Inverse(mp2);
    u:=SafeUniformiser(p);
    grp:=AbelianGroup([2:i in [1..#Invariants(grp1)+1]]);
    fn:=function(a)
	if (a eq 0) then
          error "LocalTwoSelmerMap only defined on multiplicative group";
	else
          a:=a*Denominator(a)^2;
	  v:=Valuation(a,p);
	  a:=a/u^v;
          return grp!([v]cat Eltseq(mp2(mp1(a))));
 	end if;
      end function;
    fni:=function(b)
        v:=Eltseq(b);
        return u^v[1]*mp1i(mp2i(grp2![v[i]:i in [2..#Invariants(grp1)+1]]));
      end function;
    grpmap:=pmap<NumberField(Order(p))->grp|a:->fn(a),b:->fni(b)>;
    rc:=p`Record;
    rc`LocalTwoSelmerMap:=grpmap;
    p`Record:=rc;
  end if;
  return p`Record`LocalTwoSelmerMap;
end intrinsic;

/**
 ** Support(RngOrdIdl)
 **
 ** factorisation without multiplicity.
 **/

intrinsic Support(I::RngOrdIdl) -> SetEnum
  {returns the prime divisors of an ideal as a set}
  return {p[1]:p in Factorisation(I)};
end intrinsic;

intrinsic MyFactorisation(f::RngUPolElt) -> SeqEnum
  {Stupid factorisation that only splits of linear factors. Complains if
   cofactor is of degree larger than 3}
  require LeadingCoefficient(f) eq 1 : "Can only handle monic polynomials";
  R:=Parent(f);
  if Type(BaseRing(f)) eq RngLoc then
    rts:=IntegralRoots(f);
  else
    rts:=MyRoots(f);
  end if;
  res:=[<R.1-r[1],r[2]>:r in rts];
  cofact:=&*[r[1]^r[2]:r in res];
  if Degree(f) gt Degree(cofact)+3 then
    print "WARNING: only linear factors split off";
  end if;
  return res cat [<f div cofact,1>];
end intrinsic;
